/* interreduce the Groebner basis */
INT32 ** gbasis_reduce(INT32 **a, INT32 p)
{
	INT32 *m, i, j, k;

	double c0, c1, t0, t1;
	c0 = cputime(), t0 = realtime();

	DEBUG(MATDB, "-------\n");
	a = symbol_prepro(a);
	m = polmat_matrix(a);
	a = matrix_ref(a, p, 0);
	a = polmat_poly(m, a);
	free(m);

	for (i=k=1; i < I(a[0]); i++) {
		for (j=1; j < k; j++) {
			if (monom_div(a[i][1], a[j][1])) break;
		}
		if (j==k) a[k++] = a[i];
	}
	a[0] = (INT32 *)I(k);

	c1 = cputime(), t1 = realtime();
	DEBUG(MATDB, "reduce:   %5.3f sec   %5.3f cpu\n", t1-t0, c1-c0);
	reduce_time += t1-t0;
	reduce_cput += c1-c0;

	return a;
}


INT32 ** gbasis_export()
{
	INT32 **a, i, j, k;
	a = malloc((bload+1)*sizeof(M_INT));
	for (j=1, i=0; i < bload; i++) {
		if ((M_INT)basis[i] & bfree) continue;
		a[j] = malloc(basis[i][0]*sizeof(INT32));
		for (k=0; k < basis[i][0]; k++) a[j][k] = basis[i][k];
		j++;
	}
	a[0] = (INT32 *)I(j);
	return a;
}


ALGEB gbasis_test(MKernelVector kv, ALGEB *args)
{
	ALGEB v, f, F;
	INT32 **a, p, *m, i, j, k;
	INT32 step, temp;

	double c0, c1, t0, t1;
	c0 = cputime(), t0 = realtime();

	if( LENGTH(args) != 4
		|| !IS_IMMEDIATE(args[1])	/* prime */
		|| ID(args[2]) != MAPLE_LIST	/* variables */
		|| ID(args[3]) != MAPLE_LIST	/* the basis */
	) kv->error("invalid arguments");
	cilk_init(kv);

	p = VALUE_IMMEDIATE(args[1]);
	v = A(args[2][1]);
	f = A(args[3][1]);

	table_init(v);
	basis_init();
	a = import_seq(f);

	/* invert leading coefficients */
	for (i=1; i < I(a[0]); i++) {
		matrix_normal_row(a[i], p);
	}

	update_gbasis(a);
	free(a); a = NULL;
	for (step=1; pload > 0; step++) {
		DEBUG(MATDB, "-------\n");
		DEBUG(MATDB, "STEP %d\n", step);

		// while (eload > esize/4) table_enlarge();
		DEBUG(MEMDB, "gbasis:   block=%.1f MB, table=%.1f MB\n", (double)eload*sizeof(INT32)/1024/1024, (double)tsize*sizeof(INT32)/1024/1024);

		for (i=j=0; i < bload; i++) {
			if ((M_INT)basis[i] & bfree) j++;
		}
		DEBUG(SYZDB, "gbasis:   %d/%d polynomials in basis\n", bload-j, bload);

		a = select_pairs();
		num_rowsred += I(a[0])-1;
		a = symbol_prepro(a);

		m = polmat_matrix(a);
		temp = I(a[0]);
		if (kv->getInterruptValue()) goto abort_mat;

		a = matrix_ref(a, p, 1);
		num_zerored += temp-I(a[0]);
		if (kv->getInterruptValue()) goto abort_mat;

		a = polmat_poly(m, a);
		free(m); m = NULL;

		a = symbol_filter(a);
		a = update_gbasis(a);
		free(a); a = NULL;
		if (kv->getInterruptValue()) goto abort;
	}

	a = gbasis_export();
	a = gbasis_reduce(a, p);
	f = export_seq(kv, a);
	matrix_free(a);

	c1 = cputime(), t1 = realtime();
	DEBUG(MATDB, "-------\n");
	DEBUG(BASDB, "gbasis:   %d reduced, %d eliminated, %d rows, %d zero: %3.0f%%\n", (int)num_reduced, (int)num_useless, (int)num_rowsred, (int)num_zerored, 100.*num_zerored/num_rowsred);
	DEBUG(BASDB, "gbasis:   %7.3f sec   %7.3f cpu\n", t1-t0, c1-c0);

	DEBUG(TIMDB, "select:   %7.3f sec   %7.3f cpu\n", select_time, select_cput);
	DEBUG(TIMDB, "symbol:   %7.3f sec   %7.3f cpu\n", symbol_time, symbol_cput);
	DEBUG(TIMDB, "polmat:   %7.3f sec   %7.3f cpu\n", polmat_time, polmat_cput);
	DEBUG(TIMDB, "matrix:   %7.3f sec   %7.3f cpu\n", matrix_time, matrix_cput);
	DEBUG(TIMDB, "update:   %7.3f sec   %7.3f cpu\n", update_time, update_cput);
	DEBUG(TIMDB, "reduce:   %7.3f sec   %7.3f cpu\n", reduce_time, reduce_cput);
//	DEBUG(TIMDB, "other :   %7.3f sec\n", t1 - t0 - select_time - symbol_time - polmat_time - matrix_time - update_time - reduce_time);

	basis_free();	
	table_free();
	F = kv->allocALGEB(2,MAPLE_LIST);
	F[1] = A2(f);
	return F;

	abort_mat:
		for (i=1; i < I(a[0]); i++) free(a[i]);
	abort:
		printf("computation aborted\n");
		if (a) free(a);
		if (m) free(m);
		basis_free();
		table_free();
		return NULL;
}
