/* comparison used for qsort (descending) */
static int import_cmp(const void *A, const void *B)
{
	INT32 a, b;
	a = ((INT32 *)A)[0];
	b = ((INT32 *)B)[0];
	return (int)monom_cmp(b, a);
}


/* import Maple poly dag */
static INT32 * import_poly(ALGEB a)
{
	ALGEB u;
	M_INT l, n, s, b, i, j, k, m, c;
	M_INT e[WORDSIZE/2] = {0};
	M_INT z[WORDSIZE/2] = {0};
	INT32 *E = NULL;
	INT32 *r = NULL;

	l = LENGTH(a);
	u = INDETS(a);
	n = NUMVARS(u);
	s = NUMBITS(n);
	b = BITMASK(s);
	s = FIXSHIFT(s);

	/* construct a map for u into vlist */
	/* if u[i] = v[j] then z[i-1] = j-1 */
	/* we assume u is a subset of vlist */
	for (i=1; i <= n; i++) {
		for (j=nvars; j > 0; j--) {
			if (u[i]==vlist[j]) break;
		}
		if (j==0) goto abort;
		z[i-1] = j-1;
	}

	/* exponent vector for result */
	E = malloc(nvars*sizeof(INT32));
	for (i=0; i < nvars; E[i]=0, i++);

	r = malloc((l-1)*sizeof(INT32));
	r[0] = l-1;
	for (j=1, i=POLY_DEG1; i < l; i+=2, j+=2) {
		m = UI(a[i+0]);
		c = UI(a[i+1]);
		grlex_unpack(e, m, n, s, b);
		/* re-order exponents into grevlex */
		for (k=0; k < n; k++) E[z[k]] = e[k];
		if (!IS_IMMEDIATE(c)) goto abort;
		r[j+0] = table_insert(E);
		r[j+1] = (INT32)VALUE_IMMEDIATE(c);	/* fix */
	}
	free(E);
	return r;

	abort:
	if (E) free(E);
	if (r) free(r);
	return NULL;
}


/* fill exponents and return coefficient */
static ALGEB import_term(ALGEB a, ALGEB c, INT32 *E)
{
	M_INT i, j;

	/* zero out exponent vector */
	for (i=0; i < nvars; E[i]=0, i++);

	if (IS_IMMEDIATE(a) || ID(a)==MAPLE_INTPOS || ID(a)==MAPLE_INTNEG) {
		return a;
	}
	if (ID(a)==MAPLE_PROD) {
		for (i=LENGTH(a)-2; i > 0; i-=2) {
			for (j=nvars; j > 0; j--) {
				if (a[i] != vlist[j]) continue;
				if (!IS_IMMEDIATE(a[i+1])) return NULL;
				E[j-1] = VALUE_IMMEDIATE(a[i+1]);
				break;
			}
		}
	}
	else {
		for (j=nvars; j > 0; j--) {
			if (a != A(vlist[j])) continue;
			E[j-1] = 1;
			break;
		}
	}
	return c;
}


/* import Maple sum dag */
static INT32 * import_sum(ALGEB a)
{
	ALGEB c;
	M_INT l, i, j;
	INT32 *E = NULL;
	INT32 *r = NULL;

	/* exponent vector for terms */
	E = malloc(nvars*sizeof(INT32));

	l = LENGTH(a);
	r = malloc(l*sizeof(INT32));
	r[0] = (INT32)l;
	for (i=1; i < l; i+=2) {
		c = import_term(A(a[i]), A(a[i+1]), E);
		if (!IS_IMMEDIATE(c)) goto abort;
		r[i+0] = table_insert(E);
		r[i+1] = (INT32)VALUE_IMMEDIATE(c);
	}
	free(E);
	return r;

	abort:
	if (E) free(E);
	if (r) free(r);
	return NULL;
}


/* import any Maple dag */
static INT32 * import_dag(ALGEB a)
{
	ALGEB c;
	INT32 *r, *E;

	if (ID(a)==MAPLE_POLY) {
		r = import_poly(a);
	}
	else if (ID(a)==MAPLE_SUM)
		r = import_sum(a);
	else {
		E = malloc(nvars*sizeof(INT32));
		c = import_term(a, IMMEDIATE(1), E);
		r = malloc(3*sizeof(INT32));
		r[0] = 3;
		r[1] = table_insert(E);
		r[2] = (INT32)VALUE_IMMEDIATE(c);
		free(E);
	}
	qsort( r+1, (r[0]-1)/2, 2*sizeof(INT32), &import_cmp );
	return r;
}


/* import sequence of dags */
static INT32 ** import_seq(ALGEB a)
{
	INT32 **b, i, l;

	l = LENGTH(a);
	b = malloc(l*sizeof(M_INT));
	b[0] = (INT32 *)I(l);
	for (i=1; i < l; i++) {
		b[i] = import_dag(A(a[i]));
	}
	return b;
}
