/* ==== Mega Groebner Basis library ==== */
/* Roman Pearce     rpearcea@cecm.sfu.ca */
/* Michael Monagan  mmonagan@cecm.sfu.ca */

#define MULTICORE 1	/* enable multicore using Cilk */
#define MAXTHREAD 64	/* max threads in Cilk runtime */

#define MEMDB 1		/* monomial table */
#define SYMDB 1		/* symbolic prep. */
#define POLDB 1		/* poly to matrix */
#define MATDB 1		/* matrix routine */
#define SYZDB 1		/* update s-pairs */
#define BASDB 1		/* Groebner basis */
#define TIMDB 1		/* time breakdown */

#define TABLESIZE 32768		/* initial size of hash table */

#include <stdio.h>
#include <stdlib.h>
#include <mplshlib.h>

/* statistics */
static double select_time = 0;
static double symbol_time = 0;
static double matrix_time = 0;
static double update_time = 0;
static double polmat_time = 0;
static double reduce_time = 0;

static double select_cput = 0;
static double symbol_cput = 0;
static double matrix_cput = 0;
static double update_cput = 0;
static double polmat_cput = 0;
static double reduce_cput = 0;

static M_INT  num_reduced = 0;
static M_INT  num_useless = 0;
static M_INT  num_rowsred = 0;
static M_INT  num_zerored = 0;

#include "macros.c"	/* macro definitions */
#include "system.c"	/* support functions */
#include "matrix.c"	/* sparse linear alg */

#include "hashtb.c"	/* hashing monomials */
#include "monoms.c"	/* monomial routines */
#include "import.c"	/* import Maple dags */
#include "export.c"	/* export Maple dags */
#include "polmat.c"	/* convert to matrix */

#include "global.c"	/* global vars/funcs */
#include "update.c"	/* update GB & pairs */
#include "select.c"	/* s-pairs selection */
#include "symbol.c"	/* symbolic preproc. */
#include "gbasis.c"	/* Groebner function */
