libmgb := module()
option package;
local link_external, gbasis_grid, sig, ver;
export matrix_test, import_test, export_test, polmat_test, gbasis_test, gbasis;

link_external := proc(pname)
	proc()
		try
			unprotect(procname);
			assign(procname, define_external(pname, 'MAPLE', 'LIB'=ExternalCalling:-ExternalLibraryName("mgb")));
			protect(procname);
		catch:
			error "define_external failed to link %1", pname;
		end try:
		return procname(args);
	end proc;
end proc;

matrix_test := link_external('matrix_test');
import_test := link_external('import_test');
export_test := link_external('export_test');
polmat_test := link_external('polmat_test');
gbasis_test := link_external('gbasis_test');


ver := proc() option remember;
local n;
	n := kernelopts(version);
	n := convert(n,string);
	n := StringTools:-Split(n)[-1];
	return parse(n);
end proc:

gbasis_grid := proc(p,v,F)
local n,q,i,f,g;
	n := kernelopts(numcpus);
	q[0] := p+1;
	for i to n do
		q[i] := prevprime(q[i-1]);
		f[i] := remove(`=`, F mod q[i], 0);
	end do;
	if ver() > 10^6 then	# Maple 2015
		g := [Grid:-Seq[tasksize=1](libmgb:-gbasis_test(q[i],v,f[i]), i=1..n)];
	else
		g := [Grid:-Seq(libmgb:-gbasis_test(q[i],v,f[i]), i=1..kernelopts(numcpus))];
	end if;
	for i to n do
		gbasis_grid(q[i],v,F) := g[i];
	end do;
	return g[1];
end proc:


sig := proc(f,var) local m;
	lcoeff(f,'order'='tdeg'(op(var)),'m');
	return m;
end proc;


gbasis := proc(v,F,grid::truefalse:=false)
local S,s,R,r,p,q,f,g,c,i,j,k,
TIMEG,TIMEC,TIMER,TIME1,TIME2;
	S := FAIL;
	p := 2^31;
	i := 0;
	k := 1;
	TIMEG := 0; TIMEC := 0; TIMER := 0;
	while true do
		i := i+1;
		p := prevprime(p);
		printf("%2ld: computing mod %ld\n", i, p);
		TIME1 := time[real]();

		if not grid then
			f := remove(`=`, F mod p, 0);
			g := gbasis_test(p, v, f);
		else
			g := gbasis_grid(p, v, F);
		end if;
		s := map(sig, g, v);
		TIME2 := time[real](): TIMEG := TIMEG + TIME2 - TIME1; TIME1 := TIME2;

		if s <> S then
			S := s; q := p; c := g;
		else
			c := chrem([c,g],[q,p]);
			q := q*p;
		end if;
		TIME2 := time[real](): TIMEC := TIMEC + TIME2 - TIME1; TIME1 := TIME2;

		while k <= nops(c) do
			r[k] := iratrecon(c[k],q,'scaled');
			if r[k]=FAIL then break; end if;
			k := k+1;
		end do;
		TIME2 := time[real](): TIMER := TIMER + TIME2 - TIME1; TIME1 := TIME2;
		if k > nops(c) then break; end if;
	end do;
	R := [seq(r[i]/icontent(r[i]), i=1..nops(c))];
	printf("grob: %8.3f\n", TIMEG);
	printf("chin: %8.3f\n", TIMEC);
	printf("ratr: %8.3f\n", TIMER);
	return R;
end proc:

end module:

cyclic := proc(n) global x; local i, j, k;
  [seq(add(mul(cat(x,k mod n), k=j..i+j-1), j=0..n-1), i=1..n-1), mul(cat(x,k),k=0..n-1)-1];
end proc:

katsura := proc(n) global x; local u, k, i;
  u := proc(i,n) if i < 0 then u(-i,n) elif i > n then 0 else cat(x,i) end if; end proc;
  k := proc(n,m) local i; add(u(i,n)*u(m-i,n),i=-n..n)-u(m,n) end proc:
  [seq(k(n,i), i=-n+1..0), add(u(i,n),i=-n..n)-1];
end proc:

#random := proc(n,d) global x; local u, i;
#  u := [seq(cat(x,i),i=0..n-1)];
#  [seq(randpoly(u,degree=d,dense),i=1..n)];
#end proc:

#savelib(libmgb):
#savelib(cyclic):
#savelib(katsura):
