/* monomial operations */

static void monom_print(INT32 A)
{
	INT32 i, *a;
	a = expon + A;
	printf("[");
	for (i=0; i < nvars-1; i++) {
		printf("%d,", a[i]);
	}
	printf("%d]", a[i]);
}


/* compare two monomials in grevlex */
static int monom_cmp(INT32 A, INT32 B)
{
	INT32 i, *a, *b;
	a = expon + A;
	b = expon + B;
	i = HASH_DEG;

	/* compare total degree */
	if (a[i] > b[i]) return 1;
	else if (a[i] != b[i]) return -1;

	/* compare exponents */
	for (i=nvars-1; i >= 0; i--) {
		if (a[i] < b[i]) return 1;
		else if (a[i] != b[i]) return -1;
	}
	return 0;
}


/* monomial multiplication */
static INT32 monom_mul(INT32 A, INT32 B)
{
#if 1
	INT32 *a, *b;
	a = expon + A;
	b = expon + B;
	return table_insert_prod(a, b);
#else
	INT32 *a, *b, h, i, m, *e;
	a = expon + A;
	b = expon + B;

	/* compute product and insert */
	e = alloca(nvars*sizeof(INT32));
	for (i=0; i < nvars; i++) {
		e[i] = a[i] + b[i];
	}
	m = table_insert(e);
	return m;
#endif
}


/* monomial quotient, 0 if not divide */
static INT32 monom_div(INT32 A, INT32 B)
{
	INT32 *a, *b, h, i, m, *e;
	a = expon + A;
	b = expon + B;

	/* compute quotient and insert */
	e = alloca(nvars*sizeof(INT32));
	for (i=0; i < nvars; i++) {
		e[i] = a[i] - b[i];
		if (e[i] < 0) return 0;
	}
	m = table_insert(e);
	return m;
}


/* monomial lcm */
static INT32 monom_lcm(INT32 A, INT32 B)
{
	INT32 *a, *b, *e, m, i;
	a = expon + A;
	b = expon + B;

	/* compute lcm and lookup */
	e = alloca(nvars*sizeof(INT32));
	for (i=0; i < nvars; i++) {
		e[i] = a[i] < b[i] ? b[i] : a[i];
	}
	m = table_insert(e);
	return m;
}


/* pack exponent vector e into grlex poly monomial */
/* n = number of variables, s = shift, b = bitmask */
static M_INT grlex_pack(M_INT *e, M_INT n, M_INT s, M_INT b)
{
	M_INT m, d, t;
	m = d = t = 0;
	while (n--) {
		d += e[n];
		m |= e[n] << t;
		t += s;
	}
	m |= d << t;
	return m;
}


/* unpack grlex poly monomial into exponent vector */
/* n = number of variables, s = shift, b = bitmask */
static void grlex_unpack(M_INT *e, M_INT m, M_INT n, M_INT s, M_INT b)
{
	while (n--) {
		e[n] = m & b;
		m = m >> s;
	}
}
