/* 
	collect monomials in array
	sort the monomial array
	assign indices to all monomials in array
	map monomials to indices in matrix
	do row reduction on matrix
	map indices to monomials in matrix
	assign new (pivot) rows to monomials
*/


/* comparison used for qsort (ascending) */
int polmat_cmp(const void *A, const void *B)
{
	INT32 a, b;
	a = ((INT32 *)A)[0];
	b = ((INT32 *)B)[0];
	return (int)monom_cmp(a, b);
}


/* map polynomials to matrix */
INT32 * polmat_matrix(INT32 **A)
{
	INT32 *mon, *a, *e, i, j, k;

	double c0, c1, t0, t1;
	c0 = cputime(), t0 = realtime();

	/* insert monomials into array */
	mon = malloc(eload/HASH_LEN*sizeof(INT32));
	for (k=0, i=1; i < I(A[0]); i++) {
		a = A[i];
		for (j=1; j < a[0]; j+=2) {
			e = expon + a[j];
			/* mark mon as inserted */
			if (e[HASH_IND]) continue;
			e[HASH_IND] = 1;
			mon[k++] = a[j];
 		}
	}
	mon = realloc(mon, k*sizeof(INT32));

	/* sort monomials and assign indices */
	qsort( mon, k, sizeof(INT32), &polmat_cmp );
	for (i=0; i < k; i++) {
		e = expon + mon[i];
		e[HASH_IND] = i;
	}

	/* map monomials to indices */
	for (i=1; i < I(A[0]); i++) {
		a = A[i];
		for (j=1; j < a[0]; j+=2) {
			e = expon + a[j];
			a[j] = e[HASH_IND];
		}
	}

	c1 = cputime(), t1 = realtime();
	DEBUG(POLDB, "polmat:   %5.3f sec   %5.3f cpu\n", t1-t0, c1-c0);
	polmat_time += t1-t0;
	polmat_cput += c1-c0;

	return mon;
}


/* map matrix to polynomials */
INT32 ** polmat_poly(INT32 *mon, INT32 **A)
{
	INT32 *a, *e, i, j, k;

	double c0, c1, t0, t1;
	c0 = cputime(), t0 = realtime();

	/* map indices to monomials */
	for (k=-1, i=1; i < I(A[0]); i++) {
		a = A[i];
		if (a[1] > k) k = a[1];
		for (j=1; j < a[0]; j+=2) {
			a[j] = mon[a[j]];
		}
	}

	/* remove index assignments */
	for (i=0; i <= k; i++) {
		e = expon + mon[i];
		e[HASH_IND] = 0;
	}

	c1 = cputime(), t1 = realtime();
	DEBUG(POLDB, "polmat:   %5.3f sec   %5.3f cpu\n", t1-t0, c1-c0);
	polmat_time += t1-t0;
	polmat_cput += c1-c0;

	return A;
}
