/* platform specific */

/* cilk parallelism support */
#if MULTICORE && defined(__cilk)
	#define cilk_spawn _Cilk_spawn
	#define cilk_sync  _Cilk_sync
	#define cilk_for   _Cilk_for
#else
	#undef  MULTICORE
	#define MULTICORE 0
	#define cilk_spawn
	#define cilk_sync
	#define cilk_for for
	static  int __cilkrts_get_worker_number() { return 0; }
	static  int __cilkrts_get_nworkers() { return 1; }
	static  int __cilkrts_set_param(const char* name, const char* value) { return 0; }
	static void __cilkrts_end_cilk() { return; }
#endif


void cilk_init(MKernelVector kv)
{
	M_INT n;
	char c[10];
	__cilkrts_end_cilk();
	n = VALUE_IMMEDIATE(kv->kernelopts("numcpus", NULL));
	if (n > MAXTHREAD) n = MAXTHREAD;
	printf("using %d threads\n", (int)n);
	sprintf(c, "%d\n", (int)n);
	__cilkrts_set_param("nworkers",c);
}


#include <time.h>
#include <sys/time.h>


/* cpu time */
static double cputime()
{
	double t;
	t = CLOCKS_PER_SEC/100000.;
	t += clock();
	return t/CLOCKS_PER_SEC;
}


/* wall time */
static double realtime()
{
	struct timeval t;
	gettimeofday(&t, NULL);
	/* gain precision using recent year */
	t.tv_sec -= (2010 - 1970)*3600*24*365;
	return (1. + t.tv_usec + t.tv_sec*1000000.)/1000000.;
}


/* compare and swap */
static M_INT cas(M_INT *ptr, M_INT old, M_INT new)
{
	M_INT prev;
#if MULTICORE && WORDSIZE==32
	__asm__ __volatile__("lock; cmpxchgl %2, %1" : "=a"(prev), "+m"(*ptr) : "r"(new), "0"(old) : "memory");
#elif MULTICORE && WORDSIZE==64
	__asm__ __volatile__("lock; cmpxchgq %2, %1" : "=a"(prev), "+m"(*ptr) : "r"(new), "0"(old) : "memory");
#else
	prev = *ptr;
	if (prev==old) *ptr = new;
#endif
	return prev;
}


/* xorshift PRNG */
#if WORDSIZE==64
unsigned INT64 rseed = 88172645463325252LL;
static INT64 randgen()
{
	rseed ^= (rseed << 13);
	rseed ^= (rseed >>  7);
	rseed ^= (rseed << 17);
	return rseed;
}
#elif WORDSIZE==32
unsigned INT32 rseed = 2463534242;
static INT32 randgen()
{
	rseed ^= (rseed << 13);
	rseed ^= (rseed >> 17);
	rseed ^= (rseed << 5);
	return rseed;
}
#endif


/* invert A mod p */
static INT32 modinv32(INT32 A, INT32 p)
{
	INT32 a, b, q, t, x, y;
	a = p;
	b = A % p;
	// if (b < 0) b += p;
	b += (b >> 31) & p;
	x = 1;
	y = 0;
	while (b != 0) {
		t = b;
		q = a/t;
		b = a - q*t;
		a = t;
		t = x;
		x = y - q*t;
		y = t;
	}
	// if (y < 0) y += p;
	y += (y >> 31) & p;
	return y;
}
