/* comparison used for qsort (ascending) */
static int update_cmp(const void *a, const void *b)
{
	spoly *c, *d;
	c = (spoly *)a;
	d = (spoly *)b;
	if (c->lcm != d->lcm)
		return (int)monom_cmp(c->lcm, d->lcm);
	else if (c->deg != d->deg) 
		return (c->deg < d->deg) ? -1 : 1;
	else if (c->i != d->i)
		return (c->i < d->i) ? -1 : 1;
	else
		return 0;
}


/* add polynomial to basis */
static INT32 update_insert(INT32 *r)
{
	INT32 mi, mj, lm, *b;
	INT32 i, j, k, t, e;

	/* enlarge basis */
	if (bload == bsize) {
		bsize = bsize*2;
		basis = realloc(basis, bsize*sizeof(M_INT));
	}
	/* enlarge list of pairs */
	if (pload + bload > psize) {
		psize += pload + bload;
		pairs = realloc(pairs, psize*sizeof(spoly));
	}

	/* assign new polynomial */
	basis[bload] = r;

	/* create new pairs */
	for (k=pload, i=0; i < bload; i++, k++) {
		/* basis element could be redundant */
		/* compute the lcm for criteria use */
		b = (INT32 *)((M_INT)basis[i] & bmask);
		pairs[k].i = i;
		pairs[k].j = bload;
		pairs[k].lcm = monom_lcm(b[1], r[1]);

		if (b != basis[i])
			pairs[k].deg = -1;	/* redundant */
		else if (pairs[k].lcm == monom_mul(b[1], r[1]))
			pairs[k].deg = -2;	/* criterion */
		else
			pairs[k].deg = (expon + pairs[k].lcm)[HASH_DEG];
	}
	k = bload + pload;

	/* Gebauer & Moller Bk */
	for (t=0; t < pload; t++) {
		i = pairs[t].i;
		j = pairs[t].j;
		if (pairs[t].lcm != pairs[pload+i].lcm
		 && pairs[t].lcm != pairs[pload+j].lcm
		 && monom_div(pairs[t].lcm, r[1])) {
			pairs[t].deg = -1;
		}
	}

	/* sort by increasing lcm, break ties by earlier poly */
	qsort( pairs+pload, bload, sizeof(spoly), &update_cmp );

	/* Gebauer & Moller M */
	for (i=pload; i < k; i++) {
		if (pairs[i].deg < 0) continue;
		for (j=pload; j < i; j++) {
			if (i==j || pairs[j].deg == -1) continue;	/* smaller lcm eliminated j */
			if (pairs[i].lcm != pairs[j].lcm && monom_div(pairs[i].lcm, pairs[j].lcm)) {
				pairs[i].deg = -1;
				break;
			}
		}
	}

	/* Gebauer & Moller F */
	for (i=pload; i < k; i++) {
		if (pairs[i].deg == -1) continue;
		if (pairs[i].deg == -2) {
			/* eliminate all new pairs with this lcm */
			for (j=pload; j < k; j++) {
				if (pairs[j].lcm == pairs[i].lcm)
					pairs[j].deg = -1;
			}
		}
		else {	/* earlier pairs may eliminate this pair */
			for (j=i-1; j >= pload; j--) {
				if (pairs[j].lcm == pairs[i].lcm) {
					pairs[i].deg = -1;
					break;
				}
			}
		}
	}

	/* compact pairs */
	for (e=0, i=j=0; i < k; i++) {
		if (pairs[i].deg < 0) {
			e++; continue;
		}
		pairs[j++] = pairs[i];
	}
	pload = j;

	/* remove old basis elements */
	for (i=0; i < bload; i++) {
		if ((M_INT)basis[i] & bfree) continue;
		if (monom_div(basis[i][1], r[1])) {
			basis[i] = (INT32 *)((M_INT)basis[i] | bfree);
		}
	}
	bload++;
	return e;
}


/* update basis */
static INT32 ** update_gbasis(INT32 **a)
{
	INT32 i, j, k, e;

	double c0, c1, t0, t1;
	c0 = cputime(), t0 = realtime();

	/* largest leading term first */
	k = (INT32)(M_INT)a[0];
	for (e=0, i=1; i < k; i++) {
		e += update_insert(a[k-i]);
	}
	for (i=j=0; i < bload; i++) {
		if ((M_INT)basis[i] & bfree) continue;
		j++;
	}
	c1 = cputime(), t1 = realtime();
	DEBUG(SYZDB, "update:   %d new, %d polys, %d pairs\n", k-1, j, pload);
	DEBUG(SYZDB, "update:   %5.3f sec   %5.3f cpu\n", t1-t0, c1-c0);
	update_time += t1-t0;
	update_cput += c1-c0;
	num_useless += e;

	return a;
}
